/*
 * top.v
 *
 * vim: ts=4 sw=4
 *
 * Copyright (C) 2019-2020  Sylvain Munaut <tnt@246tNt.com>
 * SPDX-License-Identifier: CERN-OHL-P-2.0
 */

`default_nettype none
`include "boards.vh"

module top (
	// SPI
	inout  wire spi_mosi,
	inout  wire spi_miso,
	inout  wire spi_clk,
	output wire spi_flash_cs_n,
`ifdef HAS_PSRAM
	output wire spi_ram_cs_n,
`endif

	// Debug UART
	input  wire uart_rx,
	output wire uart_tx,

	// Buttons (2 for now to test up and down)
	input  wire btn_n,
    input wire btn_1, // AKA write_flash
    input wire btn_2,

    // PMOD connected buttons
    input wire [15:0] pmod_btns,

    // LEDs to blink to show that a value change has been registered?
    output wire[4:0] led,

    // GPIOs for out signal
    //output wire out1, TODO: uncomment when we know the proper GPIO for that
    //output wire out2,
    //output wire out3,

	// LED TODO: remove later
	output wire [2:0] rgb,

	// Clock
	input  wire clk_in
);

	localparam integer SPRAM_AW = 14; /* 14 => 64k, 15 => 128k */
	localparam integer WB_N  =  7;

	localparam integer WB_DW = 32;
	localparam integer WB_AW = 16;
	localparam integer WB_RW = WB_DW * WB_N;
	localparam integer WB_MW = WB_DW / 8;

    localparam integer FAST_PWM_WIDTH = 8;
    localparam integer PULSE_COUNTER_WIDTH = 8;
    localparam integer SLOW_PWM_WIDTH = 14;

    localparam integer BUTTON_COUNTER_WIDTH = 8; //TODO - zmienić na 8

	genvar i;


	// Signals
	// -------

	// Wishbone
	wire [WB_AW-1:0] wb_addr;
	wire [WB_DW-1:0] wb_rdata [0:WB_N-1];
	wire [WB_RW-1:0] wb_rdata_flat;
	wire [WB_DW-1:0] wb_wdata;
	wire [WB_MW-1:0] wb_wmsk;
	wire [WB_N -1:0] wb_cyc;
	wire             wb_we;
	wire [WB_N -1:0] wb_ack;

	// WarmBoot
	reg boot_now;
	reg [1:0] boot_sel;

	// Clock / Reset logic
	wire clk_24_75m;
	wire clk_99m;
	wire rst;

    // 3 signal
    wire [SLOW_PWM_WIDTH-1:0] period1;
    wire [SLOW_PWM_WIDTH-1:0] delay1;

    wire [SLOW_PWM_WIDTH-1:0] period2;
    wire [SLOW_PWM_WIDTH-1:0] delay2;
    
	wire [FAST_PWM_WIDTH-1:0] period3;
    wire [FAST_PWM_WIDTH-1:0] duty3;
    wire [SLOW_PWM_WIDTH-1:0] delay3;
    wire [PULSE_COUNTER_WIDTH-1:0] npuls3;
    
	wire [1:0] odd_train_flag;
    
	wire ena_odd_out3;
	assign ena_odd_out3=1; //TODO

    // Buttons
    reg [BUTTON_COUNTER_WIDTH-1:0] inc_f1;
    reg [BUTTON_COUNTER_WIDTH-1:0] dec_f1;
    reg [BUTTON_COUNTER_WIDTH-1:0] inc_d1;
    reg [BUTTON_COUNTER_WIDTH-1:0] dec_d1;
    reg [BUTTON_COUNTER_WIDTH-1:0] inc_d2;
    reg [BUTTON_COUNTER_WIDTH-1:0] dec_d2;
    reg [BUTTON_COUNTER_WIDTH-1:0] inc_ph2;
    reg [BUTTON_COUNTER_WIDTH-1:0] dec_ph2;
    reg [BUTTON_COUNTER_WIDTH-1:0] inc_f3;
    reg [BUTTON_COUNTER_WIDTH-1:0] dec_f3;
    reg [BUTTON_COUNTER_WIDTH-1:0] inc_d3;
    reg [BUTTON_COUNTER_WIDTH-1:0] dec_d3;
    reg [BUTTON_COUNTER_WIDTH-1:0] inc_ph3;
    reg [BUTTON_COUNTER_WIDTH-1:0] dec_ph3;
    reg [BUTTON_COUNTER_WIDTH-1:0] inc_n3;
    reg [BUTTON_COUNTER_WIDTH-1:0] dec_n3;

	// Mailbox signal wires
    wire [16*16-1:0] mailbox_regs_flat;  // Flattened register array (16 registers of 16 bits each)
    wire [BUTTON_COUNTER_WIDTH*16-1:0] mailbox_btns_flat;

	// SoC
	// ---

	soc_picorv32_base #(
		.WB_N    (WB_N),
		.WB_DW   (WB_DW),
		.WB_AW   (WB_AW),
		.SPRAM_AW(SPRAM_AW)
	) base_I (
		.wb_addr (wb_addr),
		.wb_rdata(wb_rdata_flat),
		.wb_wdata(wb_wdata),
		.wb_wmsk (wb_wmsk),
		.wb_we   (wb_we),
		.wb_cyc  (wb_cyc),
		.wb_ack  (wb_ack),
		.clk     (clk_24_75m),
		.rst     (rst)
	);

	for (i=0; i<WB_N; i=i+1)
		assign wb_rdata_flat[i*WB_DW+:WB_DW] = wb_rdata[i];


	// UART [1]
	// ----

	uart_wb #(
		.DIV_WIDTH(12),
		.DW(WB_DW)
	) uart_I (
		.uart_tx  (uart_tx),
		.uart_rx  (uart_rx),
		.wb_addr  (wb_addr[1:0]),
		.wb_rdata (wb_rdata[1]),
		.wb_we    (wb_we),
		.wb_wdata (wb_wdata),
		.wb_cyc   (wb_cyc[1]),
		.wb_ack   (wb_ack[1]),
		.clk      (clk_24_75m),
		.rst      (rst)
	);


	// SPI [2]
	// ---

	ice40_spi_wb #(
`ifdef HAS_PSRAM
		.N_CS(2),
`else
		.N_CS(1),
`endif
		.WITH_IOB(1),
		.UNIT(0)
	) spi_I (
		.pad_mosi (spi_mosi),
		.pad_miso (spi_miso),
		.pad_clk  (spi_clk),
`ifdef HAS_PSRAM
		.pad_csn  ({spi_ram_cs_n, spi_flash_cs_n}),
`else
		.pad_csn  (spi_flash_cs_n),
`endif
		.wb_addr  (wb_addr[3:0]),
		.wb_rdata (wb_rdata[2]),
		.wb_wdata (wb_wdata),
		.wb_we    (wb_we),
		.wb_cyc   (wb_cyc[2]),
		.wb_ack   (wb_ack[2]),
		.clk      (clk_24_75m),
		.rst      (rst)
	);


	// RGB LEDs [3]
	// --------
	assign wb_ack[3]=wb_cyc[3];

	
	//ice40_rgb_wb #(
	//	.CURRENT_MODE("0b1"),
	//	.RGB0_CURRENT("0b000001"),
	//	.RGB1_CURRENT("0b000001"),
	//	.RGB2_CURRENT("0b000001")
	//) rgb_I (
	//	.pad_rgb    (rgb),
	//	.wb_addr    (wb_addr[4:0]),
	//	.wb_rdata   (wb_rdata[3]),
	//	.wb_wdata   (wb_wdata),
	//	.wb_we      (wb_we),
	//	.wb_cyc     (wb_cyc[3]),
	//	.wb_ack     (wb_ack[3]),
	//	.clk        (clk_24_75m),
	//	.rst        (rst)
	//);

	// WB Register Mailbox [4]
	// ----------
	mailbox_wb_sw2rtl #(
		.AW(4),
		.DW(WB_DW)
	) mailbox_regs_I (
		.clk(clk_24_75m),
		.rst(rst),
		.wb_addr(wb_addr[3:0]),
		.wb_wdata(wb_wdata),
		.wb_rdata(wb_rdata[4]),
		.wb_we(wb_we),
		.wb_cyc(wb_cyc[4]),
		.wb_ack(wb_ack[4]),
        .write_flash(btn_1),
		.registers_flat(mailbox_regs_flat)
	);

    // 3 Signal
    // --------
    assign period1 = mailbox_regs_flat[15:0];        // First 16 bits for period1
    assign delay1 = mailbox_regs_flat[31:16];        // Next 16 bits for delay1
    
	assign period2 = mailbox_regs_flat[47:32];       // Next 16 bits for period2
    assign delay2 = mailbox_regs_flat[63:48];        // Next 16 bits for delay2
    
	assign period3 = mailbox_regs_flat[79:64];       // Next 16 bits for period3
    assign duty3 = mailbox_regs_flat[95:80];         // Next 16 bits for duty3
    assign delay3 = mailbox_regs_flat[111:96];       // Next 16 bits for delay3
    assign npuls3 = mailbox_regs_flat[127:112];      // Next 16 bits for npuls3
    
	assign odd_train_flag = mailbox_regs_flat[143:128]; // Next 16 bits for odd_train_flag

    three_signal #(
        .FAST_PWM_WIDTH(FAST_PWM_WIDTH),
        .PULSE_COUNTER_WIDTH(PULSE_COUNTER_WIDTH),
        .SLOW_PWM_WIDTH(SLOW_PWM_WIDTH)
    ) three_signal_I(
        .nrst(~rst),
        .clk(clk_99m), // TODO: fix later we have CDC problem here
        .period1(period1),
        .delay1(delay1),
        .period2(period2),
        .delay2(delay2),
        .period3(period3),
        .duty3(duty3),
        .delay3(delay3),
        .npuls3(npuls3),
        .odd_train_flag(odd_train_flag),
        .ena_odd_out3(ena_odd_out3),
        .Out1(led[1]), //led[1] - P2_1
        .Out2(led[2]), //led[2] - P2_2
        .Out3(led[4]) //led[4] - P2_3
    );

	// WB Button Mailbox [5]
	// ----------
	mailbox_wb_rtl2sw #(
		.AW(4),
		.DW(WB_DW)
	) mailbox_btns_I (
		.clk(clk_24_75m),
		.rst(rst),
		.wb_addr(wb_addr[3:0]),
		.wb_rdata(wb_rdata[5]),
		.wb_cyc(wb_cyc[5]),
		.wb_ack(wb_ack[5]),
		.registers_flat_in(mailbox_btns_flat)
	);

    // WB Simple timer [6]
    timer_wb #(
        .DW(WB_DW)
    ) timer_I (
        .clk(clk_24_75m),
        .rst(rst),
        .wb_rdata(wb_rdata[6]),
        .wb_cyc(wb_cyc[6]),
        .wb_ack(wb_ack[6])
    );

    // Buttons
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(0  + 1) - 1:BUTTON_COUNTER_WIDTH*0 ] = inc_f1;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(1  + 1) - 1:BUTTON_COUNTER_WIDTH*1 ] = dec_f1;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(2  + 1) - 1:BUTTON_COUNTER_WIDTH*2 ] = inc_d1;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(3  + 1) - 1:BUTTON_COUNTER_WIDTH*3 ] = dec_d1;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(4  + 1) - 1:BUTTON_COUNTER_WIDTH*4 ] = inc_d2;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(5  + 1) - 1:BUTTON_COUNTER_WIDTH*5 ] = dec_d2;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(6  + 1) - 1:BUTTON_COUNTER_WIDTH*6 ] = inc_ph2;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(7  + 1) - 1:BUTTON_COUNTER_WIDTH*7 ] = dec_ph2;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(8  + 1) - 1:BUTTON_COUNTER_WIDTH*8 ] = inc_f3;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(9  + 1) - 1:BUTTON_COUNTER_WIDTH*9 ] = dec_f3;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(10 + 1) - 1:BUTTON_COUNTER_WIDTH*10] = inc_d3;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(11 + 1) - 1:BUTTON_COUNTER_WIDTH*11] = dec_d3;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(12 + 1) - 1:BUTTON_COUNTER_WIDTH*12] = inc_ph3;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(13 + 1) - 1:BUTTON_COUNTER_WIDTH*13] = dec_ph3;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(14 + 1) - 1:BUTTON_COUNTER_WIDTH*14] = inc_n3;
    assign mailbox_btns_flat[BUTTON_COUNTER_WIDTH*(15 + 1) - 1:BUTTON_COUNTER_WIDTH*15] = dec_n3;

    /*assign mailbox_btns_flat[15:0] =    16'd666; 
    assign mailbox_btns_flat[31:16] =   16'd1000+rst; 
    assign mailbox_btns_flat[47:32] =   pmod_btns[2]; 
    assign mailbox_btns_flat[63:48] =   pmod_btns[3]; 
    assign mailbox_btns_flat[79:64] =   pmod_btns[4]; 
    assign mailbox_btns_flat[95:80] =   pmod_btns[5]; 
    assign mailbox_btns_flat[111:96] =  pmod_btns[6]; 
    assign mailbox_btns_flat[127:112] = pmod_btns[7]; 
    assign mailbox_btns_flat[143:128] = pmod_btns[8]; 
    assign mailbox_btns_flat[159:144] = pmod_btns[9]; 
    assign mailbox_btns_flat[175:160] = pmod_btns[10];
    assign mailbox_btns_flat[191:176] = pmod_btns[11];
    assign mailbox_btns_flat[207:192] = pmod_btns[12];
    assign mailbox_btns_flat[223:208] = pmod_btns[13];
    assign mailbox_btns_flat[239:224] = pmod_btns[14];
    assign mailbox_btns_flat[255:240] = pmod_btns[15];*/

	reg [0:0] tick_1kHz, tick_100Hz;
	reg [15:0] cnt_1kHz;
	reg [4:0] cnt_100Hz;
	always @(posedge clk_24_75m or posedge rst) begin
		if(rst) begin
			tick_1kHz<=0;
			cnt_1kHz<=0;
		end else begin
			tick_1kHz<=0;
			tick_100Hz<=0;

			if (cnt_1kHz==0) begin
				tick_1kHz<=1;
				cnt_1kHz<=16'd24750;
			end else begin
				cnt_1kHz<=cnt_1kHz-1;
			end

			if(tick_1kHz) begin
				if(cnt_100Hz==0) begin
					cnt_100Hz<=9;
					tick_100Hz<=1;
				end else begin
					cnt_100Hz<=cnt_100Hz-1;
				end
			end
		end
	end

	assign led[0] = ~pmod_btns[0];

	button b1(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[0]),
		.press_count(inc_f1)

	);

	button b2(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[1]),
		.press_count(dec_f1)
	);

	button b3(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[2]),
		.press_count(inc_d1)
	);

	button b4(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[3]),
		.press_count(dec_d1)
	);

	button b5(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[4]),
		.press_count(inc_d2)
	);

	button b6(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[5]),
		.press_count(dec_d2)
	);

	button b7(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[6]),
		.press_count(inc_ph2)
	);

	button b8(
		.clk(clk_24_75m),
		.nrst(~rst),	
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[7]),
		.press_count(dec_ph2)
	);

	button b9(
		.clk(clk_24_75m),
		.nrst(~rst),		
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[8]),
		.press_count(inc_f3)
	);

	button b10(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[9]),
		.press_count(dec_f3)
	);

	button b11(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[10]),
		.press_count(inc_d3)
	);

	button b12(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[11]),
		.press_count(dec_d3)
	);

	button b13(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[12]),
		.press_count(inc_ph3)
	);

	button b14(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[13]),
		.press_count(dec_ph3)
	);

	button b15(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[14]),
		.press_count(inc_n3)
	);

	button b16(
		.clk(clk_24_75m),
		.nrst(~rst),
		.debounce_tick(tick_1kHz),
  		.cnt_tick(tick_100Hz),
		.butt(~pmod_btns[15]),
		.press_count(dec_n3)
	);

    // TODO: leaving until testing is done - btn pres will assert if RTL is
    // running still
    always @(posedge clk_24_75m or posedge rst)
        if (rst) begin
            led[3] = 1'b0;
        end else if (btn_2) begin
            led[3] = 1'b1;
        end else begin
            led[3] = 1'b0;
        end


	// Warm Boot
	// ---------

	// Bus interface
	always @(posedge clk_24_75m or posedge rst)
		if (rst) begin
			boot_now <= 1'b0;
			boot_sel <= 2'b00;
		end else if (wb_cyc[0] & wb_we & (wb_addr[2:0] == 3'b000)) begin
			boot_now <= wb_wdata[2];
			boot_sel <= wb_wdata[1:0];
		end

	assign wb_rdata[0] = 0;
	assign wb_ack[0] = wb_cyc[0];

	// Helper
	dfu_helper #(
		.TIMER_WIDTH(24),
		.BTN_MODE(3),
		.DFU_MODE(0)
	) dfu_helper_I (
		.boot_now(boot_now),
		.boot_sel(boot_sel),
		.btn_pad(btn_n),
		.btn_val(),
		.rst_req(),
		.clk(clk_24_75m),
		.rst(rst)
	);


	// Clock / Reset
	// -------------

`ifdef SIM
	reg clk_99m_s = 1'b0;
	reg clk_24_75m_s = 1'b0;
	reg rst_s = 1'b1;

	always #5 clk_99m_s <= !clk_99m_s;
	always #20 clk_24_75m_s <= !clk_24_75m_s;

	initial begin
		#200 rst_s = 0;
	end

	assign clk_99m = clk_99m_s;
	assign clk_24_75m = clk_24_75m_s;
	assign rst = rst_s;
`else
	sysmgr sys_mgr_I (
		.clk_in(clk_in),
		.rst_in(1'b0),
		.clk_99m(clk_99m),
		.clk_24_75m(clk_24_75m),
		.rst_out(rst)
	);
`endif

endmodule // top
